from pathlib import Path
from typing import Optional

import pandas as pd

from .commons import download_data

""" tools to download and load data from JHU's CSSE Covid tracker """

CSSE_REPO_BASE_URL = "https://raw.githubusercontent.com/CSSEGISandData/COVID-19/master/csse_covid_19_data/csse_covid_19_daily_reports/"
DATE_FMT = "%m-%d-%Y"

DROP_SCHEMA_V1 = ["FIPS", "Admin2", "Last_Update", "Lat", "Long_", "Combined_Key", "Incidence_Rate", "Case-Fatality_Ratio", "Country_Region"]
DROP_SCHEMA_V2 = ["FIPS", "Admin2", "Last_Update", "Lat", "Long_", "Combined_Key", "Incident_Rate",  "Case_Fatality_Ratio", "Country_Region"]

def fetch(dst: Path, date: pd.Timestamp, overwrite: bool = False) -> None:
    filename = date.strftime(DATE_FMT) + ".csv"
    if (not (dst/filename).exists()) or overwrite:
        download_data(dst, filename, CSSE_REPO_BASE_URL)

def fetch_range(dst: Path, start: str, end: str):
    for date in pd.date_range(pd.Timestamp(start), pd.Timestamp(end)):
        fetch(dst, date)

def load(dst: Path, start: str, end: str, selector: Optional[str] = None) -> pd.DataFrame:
    return pd.concat([
        (lambda _: _.query(selector) if selector else _)(pd.read_csv(dst/(date.strftime(DATE_FMT) + ".csv"))).assign(date = date) for date in pd.date_range(pd.Timestamp(start), pd.Timestamp(end))
    ], axis = 0)

def load_country(dst: Path, start: str, end: str, country: str, schema_version: int = 1):
    return load(dst, start, end, f"Country_Region == '{country}'")\
        .drop(columns = DROP_SCHEMA_V1 if schema_version == 1 else DROP_SCHEMA_V2)\
        .assign(Active = lambda _:_["Active"].astype(int))

def assemble_timeseries(df: pd.DataFrame, province: Optional[str] = None):
    totals = (
        df[df.Province_State == province].set_index("date") if province else 
        df.set_index(["date", "Province_State"]).stack().sum(level = [0, 2]).unstack()
    )[["Deaths", "Recovered", "Confirmed"]]\
        .rename(columns = {"Confirmed": "T", "Deaths": "D", "Recovered": "R"})
    return pd.concat([
        totals, 
        totals.diff()\
            .rename(lambda x: "d" + x, axis = 1)
        ], axis = 1)\
            .dropna()\
            .astype(int)


from pathlib import Path
import requests


def download_data(data_path: Path, filename: str, base_url: str = 'https://api.covid19india.org/csv/latest/'):
    """ download a file with filename from the base_url to the directory at data_path  """
    url = base_url + filename
    response = requests.get(url)
    with (data_path/filename).open('wb') as dst:
        dst.write(response.content)



from abc import abstractmethod
from itertools import product
from typing import Dict, List, Optional, Tuple

import numpy as np
from sklearn.metrics import auc

from scipy.stats import multinomial as Multinomial

from .models import SIR
from .utils import weeks

def AUC(curve):
    return auc(x = range(len(curve)), y = curve)

# Adaptive Control policies
def simulate_lockdown(model: SIR, lockdown_period: int, total_time: int, Rt0_mandatory: Dict[str, float], Rt0_voluntary: Dict[str, float], lockdown: np.matrix, migrations: np.matrix) -> SIR:
    """ simulate a lockdown where stringency is assumed to be reflected by Rt0_mandatory, with a return to Rt0_voluntary (and possible migration) once lockdown is over """
    return model.set_parameters(Rt0 = Rt0_mandatory)\
        .run(lockdown_period,  migrations = lockdown)\
        .set_parameters(Rt0 = Rt0_voluntary)\
        .run(total_time - lockdown_period, migrations = migrations)

def simulate_adaptive_control(
    model: SIR, 
    initial_run: int, 
    total_time: int, 
    lockdown: np.matrix, 
    migrations: np.matrix, 
    R_m: Dict[str, float],
    beta_v: Dict[str, float], 
    beta_m: Dict[str, float], 
    evaluation_period: int = 2*weeks, 
    adjacency: Optional[np.matrix] = None) -> SIR:
    """ simulate the Malani et al. adaptive lockdown policy where districts are assigned to lockdown stringency buckets based on Rt """
    n = len(model)
    model.set_parameters(Rt0 = R_m)\
         .run(initial_run, lockdown)
    days_run = initial_run
    gantt = []
    last_category = dict()
    while days_run < total_time:
        Gs, Ys, Os, Rs = set(), set(), set(), set()
        categories = dict(enumerate([Gs, Ys, Os, Rs]))
        category_transitions = {}
        for (i, unit) in enumerate(model):
            latest_Rt = unit.Rt[-1]
            if latest_Rt < 1: 
                Gs.add(i)
                beta_cat = 0
            else: 
                if days_run < initial_run + evaluation_period: # force first period to be lockdown
                    Rs.add(i)
                    beta_cat = 3
                else: # categorize districts based on Rt 
                    if latest_Rt < 1.5: 
                        Ys.add(i)
                        beta_cat = 1
                    elif latest_Rt < 2: 
                        Os.add(i)
                        beta_cat = 2
                    else:
                        Rs.add(i)
                        beta_cat = 3
            if unit.name not in last_category:
                last_category[unit.name] = beta_cat
            else: 
                old_beta_cat = last_category[unit.name]
                if old_beta_cat != beta_cat:
                    if beta_cat < old_beta_cat and beta_cat != (old_beta_cat - 1): # force gradual release
                        beta_cat = old_beta_cat - 1
                        if i in categories[old_beta_cat]: categories[old_beta_cat].remove(i)
                        categories[beta_cat].add(i)
                    category_transitions[unit.name] = beta_cat
                    last_category[unit.name] = beta_cat 
            gantt.append([unit.name, days_run, beta_cat, max(0, latest_Rt)])

        for (unit_name, beta_cat) in category_transitions.items(): 
            unit =  model[unit_name]
            new_beta = beta_v[unit.name] - (beta_cat * (beta_v[unit.name] - beta_m[unit.name])/3.0)                
            unit.beta[-1] = new_beta
            unit.Rt0 = new_beta * unit.gamma

        phased_migration = migrations.copy()
        for (i, j) in product(range(n), range(n)):
            if i not in Gs or j not in Gs:
                phased_migration[i, j] = 0
        model.run(evaluation_period, phased_migration)
        days_run += evaluation_period

    model.gantt = gantt 
    return model 

def simulate_adaptive_control_MHA(model: SIR, initial_run: int, total_time: int, lockdown: np.matrix, migrations: np.matrix, R_m: Dict[str, float], beta_v: Dict[str, float], beta_m: Dict[str, float], evaluation_period = 2*weeks):
    """ simulates the version of adaptive control suggested by the Indian Ministry of Home Affairs, where the trigger is based on infection count doubling time """
    n = len(model)
    model.set_parameters(Rt0 = R_m)\
         .run(initial_run, lockdown)
    days_run = initial_run
    gantt = []
    last_category = dict()
    while days_run < total_time:
        Gs, Ys, Os, Rs = set(), set(), set(), set()
        categories = dict(enumerate([Gs, Ys, Os, Rs]))
        category_transitions = {}
        for (i, unit) in enumerate(model):
            latest_Rt = unit.Rt[-1]
            if days_run < initial_run + evaluation_period: # force first period to MHA
                if unit.I[-4] != 0 and unit.I[-1]/unit.I[-4] > 2: # doubling time trigger 
                    Rs.add(i)
                    beta_cat = 3
                else:
                    Gs.add(i)
                    beta_cat = 0
            else: 
                if latest_Rt < 1: 
                    Gs.add(i)
                    beta_cat = 0
                elif latest_Rt < 1.5: 
                    Ys.add(i)
                    beta_cat = 1
                elif latest_Rt < 2: 
                    Os.add(i)
                    beta_cat = 2
                else:
                    Rs.add(i)
                    beta_cat = 3
            if unit.name not in last_category:
                last_category[unit.name] = beta_cat
            else: 
                old_beta_cat = last_category[unit.name]
                if old_beta_cat != beta_cat:
                    if beta_cat < old_beta_cat and beta_cat != (old_beta_cat - 1): # force gradual release
                        beta_cat = old_beta_cat - 1
                    if i in categories[old_beta_cat]: categories[old_beta_cat].remove(i)
                    categories[beta_cat].add(i)
                    category_transitions[unit.name] = beta_cat
                    last_category[unit.name] = beta_cat 
            gantt.append([unit.name, days_run, beta_cat, max(0, latest_Rt)])

        for (unit_name, beta_cat) in category_transitions.items(): 
            unit =  model[unit_name]
            new_beta = beta_v[unit.name] - (beta_cat * (beta_v[unit.name] - beta_m[unit.name])/3.0)                
            unit.beta[-1] = new_beta
            unit.Rt0 = new_beta * unit.gamma

        phased_migration = migrations.copy()
        for (i, j) in product(range(n), range(n)):
            if i not in Gs or j not in Gs:
                phased_migration[i, j] = 0
        model.run(evaluation_period, phased_migration)
        days_run += evaluation_period

    model.gantt = gantt 
    return model 

def simulate_PID_controller(
    model: SIR, 
    initial_run: int, 
    total_time: int,
    Rtarget: float = 0.9,
    kP: float = 0.05, 
    kI: float = 0.5,
    kD: float = 0,
    Dt: float = 1.0) -> SIR:
    """ implements a hypothetical proportional-integral-derivative control policy where the error term is Rt magnitude above 1 """
    # initial run without PID 
    model.run(initial_run)
    
    # set up PID running variables
    integral   = 0
    derivative = 0
    u = 0
    prev_error = model[0].Rt[-1]

    z = np.zeros((len(model), len(model)))

    # run forward model 
    for i in range(total_time - initial_run):
        model[0].Rt0 -= u 
        model.run(1, z)

        error = model[0].Rt[-1] - Rtarget
        integral  += error * Dt 
        derivative = (error - prev_error)/Dt

        u = kP * error + kI * integral + kD * derivative
        prev_error = error
    return model 

# Vaccination policies
class VaccinationPolicy():
    """ parent class to hold vaccination policy state """
    def __init__(self, bin_populations: np.array) -> None:
        self.bin_populations = bin_populations

    def name(self) -> str:
        return self.__class__.__name__.lower()

    @abstractmethod
    def distribute_doses(self, model: SIR, *kwargs) -> Tuple[np.array]:
        pass 

    def exhausted(self, model) -> bool:
        return self.daily_doses * len(model.Rt) > model.pop0

    def get_mortality(self, base_IFRs) -> float:
        """ update mortality based on composition of I compartment """
        if self.bin_populations.sum() == 0:
            return 0
        return base_IFRs @ self.bin_populations/self.bin_populations.sum()

class RandomVaccineAssignment(VaccinationPolicy):
    """ assigns vaccines to members of the population randomly """
    def __init__(self, daily_doses: int, effectiveness: float, bin_populations: np.array, age_ratios: np.array):
        self.daily_doses = daily_doses 
        self.age_ratios = age_ratios
        self.effectiveness = effectiveness
        self.bin_populations = bin_populations

    def distribute_doses(self, model: SIR, num_sims: int = 10000) -> Tuple[np.array]:
        if self.exhausted(model):
            return (np.zeros(self.age_ratios.shape), np.zeros(self.age_ratios.shape), np.zeros(self.age_ratios.shape))
        dV = (model.S[-1]/model.N[-1]) * self.daily_doses * self.effectiveness
        model.S[-1] -= dV
        model.parallel_forward_epi_step(num_sims = num_sims)
        distributed_doses = Multinomial.rvs(self.daily_doses, self.age_ratios)
        effective_doses   = self.effectiveness * distributed_doses
        immunizing_doses  = (model.S[-1].mean()/model.N[-1].mean()) * effective_doses
        self.bin_populations -= immunizing_doses.astype(int)
        return (distributed_doses, effective_doses, immunizing_doses)

    def name(self) -> str:
        return "randomassignment"

class PrioritizedAssignment(VaccinationPolicy):
    """ assigns vaccines to members of the population based on a prioritized ordering of subpopulations """
    def __init__(self, daily_doses: int, effectiveness: float, bin_populations: np.array, prioritization: List[int], label: str):
        self.daily_doses     = daily_doses
        self.bin_populations = bin_populations
        self.prioritization  = prioritization
        self.effectiveness   = effectiveness
        self.label = label

    def name(self) -> str:
        return f"{self.label}prioritized"

    def distribute_doses(self, model: SIR, num_sims: int = 10_000) -> Tuple[np.array]:
        if self.exhausted(model):
            return (None, None, None)
            # return (np.zeros(self.age_ratios.shape), np.zeros(self.age_ratios.shape), np.zeros(self.age_ratios.shape))
        dV = (model.S[-1]/model.N[-1]) * self.daily_doses * self.effectiveness
        model.S[-1] -= dV
        model.parallel_forward_epi_step(num_sims = num_sims)

        dVx = np.zeros(self.bin_populations.shape)
        bin_idx, age_bin = next(((i, age_bin) for (i, age_bin) in enumerate(self.prioritization) if self.bin_populations[age_bin] > 0), (None, None))
        if age_bin is not None:
            if self.bin_populations[age_bin] > self.daily_doses:
                self.bin_populations[age_bin] -= self.daily_doses
                dVx[age_bin] = self.daily_doses
            else: 
                leftover = self.daily_doses - self.bin_populations[age_bin]
                dVx[age_bin] = self.bin_populations[age_bin]
                self.bin_populations[age_bin] = 0
                if bin_idx != len(self.bin_populations) - 1:
                    dVx[self.prioritization[bin_idx + 1]] = leftover
                    self.bin_populations[self.prioritization[bin_idx + 1]] -= leftover 
        else: 
            print("vaccination exhausted", self.bin_populations, self.prioritization)
        return (
            dVx, 
            dVx * self.effectiveness, 
            dVx * self.effectiveness * (model.S[-1].mean()/model.N[-1].mean())
        )



from pathlib import Path
from typing import Dict, Sequence

import numpy as np
import pandas as pd

""" set up migration matrices from DDL matrix data """

district_2011_replacements = {
    'Maharashtra' : {
        'Mumbai Suburban' : 'Mumbai'}
 }

def load_migration_matrix(matrix_path: Path, populations: np.array) -> np.matrix:
    M  = np.loadtxt(matrix_path, delimiter=',') # read in raw data
    M *= populations[:,  None]                  # weight by population
    M /= M.sum(axis = 0)                        # normalize
    return M

def district_migration_matrices(
    matrix_path: Path, 
    states: Sequence[str]) -> Dict[str, np.matrix]:
    mm = pd.read_csv(matrix_path)
    aggregations = dict()
    for col in  ['D_StateCensus2011', 'D_DistrictCensus2011', 'O_StateCensus2011', 'O_DistrictCensus2011']:
        mm[col] = mm[col].str.title().str.replace("&", "and")
    for state in  states:
        mm_state = mm[(mm.D_StateCensus2011 == state) & (mm.O_StateCensus2011 == state)]
        # handle states that need migration data combined (e.g. Mumbai and Mumbai Suburban)
        if state in district_2011_replacements:
            mm_state.replace(district_2011_replacements[state], inplace=True)
        # group to combine multiple districts with same name based on above
        grouped_mm_state = mm_state.groupby(['D_DistrictCensus2011', 'O_DistrictCensus2011'])[['O_Population_2011','NSS_STMigrants']].sum().reset_index()
        pivot = grouped_mm_state.pivot(index = "D_DistrictCensus2011", columns = "O_DistrictCensus2011", values = "NSS_STMigrants").fillna(0)
        M  = np.matrix(pivot)
        Mn = M/M.sum(axis = 0)
        Mn[np.isnan(Mn)] = 0
        aggregations[state] = (
            pivot.index, 
            grouped_mm_state.groupby("O_DistrictCensus2011")["O_Population_2011"].agg(lambda x: list(x)[0]).values, 
            Mn
        )
    return aggregations


import logging
from typing import Callable, Optional, Sequence

import arviz as az
import numpy as np
import pandas as pd
import pymc3 as pm
import theano.tensor as tt
from scipy.stats import gamma as Gamma
from scipy.stats import nbinom
from statsmodels.regression.linear_model import OLS
from statsmodels.regression.rolling import RollingOLS
from statsmodels.tools import add_constant

from .utils import days

logger = logging.getLogger(__name__)

def rollingOLS(totals: pd.DataFrame, window: int = 3, infectious_period: float = 4.5) -> pd.DataFrame:
    """ legacy rolling regression-based implementation of Bettencourt/Ribeiro method """
    # run rolling regressions and get parameters
    model   = RollingOLS.from_formula(formula = "logdelta ~ time", window = window, data = totals)
    rolling = model.fit(method = "lstsq")
    
    growthrates = rolling.params.join(rolling.bse, rsuffix="_stderr")
    growthrates["rsq"] = rolling.rsquared
    growthrates.rename(lambda s: s.replace("time", "gradient").replace("const", "intercept"), axis = 1, inplace = True)

    # calculate growth rates
    growthrates["egrowthrateM"] = growthrates.gradient + 2 * growthrates.gradient_stderr
    growthrates["egrowthratem"] = growthrates.gradient - 2 * growthrates.gradient_stderr
    growthrates["R"]            = growthrates.gradient * infectious_period + 1
    growthrates["RM"]           = growthrates.gradient + 2 * growthrates.gradient_stderr * infectious_period + 1
    growthrates["Rm"]           = growthrates.gradient - 2 * growthrates.gradient_stderr * infectious_period + 1
    growthrates["date"]         = growthrates.index.get_level_values('status_change_date')
    growthrates["days"]         = totals.time

    return growthrates

def analytical_MPVS(
        infection_ts: pd.DataFrame, 
        smoothing: Callable,
        alpha: float = 3.0,                # shape 
        beta:  float = 2.0,                # rate
        CI:    float = 0.95,               # confidence interval 
        infectious_period: int = 5*days,   # inf period = 1/gamma,
        variance_shift: float = 0.99,      # how much to scale variance parameters by when anomaly detected 
        totals: bool = True                # are these case totals or daily new cases?
    ):
    """Estimates Rt ~ Gamma(alpha, 1/beta), and implements an analytical expression for a mean-preserving variance increase whenever case counts fall outside the CI defined by a negative binomial distribution"""
    # infection_ts = infection_ts.copy(deep = True)
    dates = infection_ts.index
    if totals:
        # daily_cases = np.diff(infection_ts.clip(lower = 0)).clip(min = 0) # infection_ts clipped because COVID19India API does weird stuff
        daily_cases = infection_ts.clip(lower = 0).diff().clip(lower = 0).iloc[1:]
    else: 
        daily_cases = infection_ts 
    total_cases = np.cumsum(smoothing(np.squeeze(daily_cases)))

    v_alpha, v_beta = [], []

    Rt_pred, Rt_CI_upper, Rt_CI_lower = [], [], []

    T_pred, T_CI_upper, T_CI_lower = [], [], []

    new_cases_ts = []

    anomalies     = []
    anomaly_dates = []

    for i in range(2, len(total_cases)):
        new_cases     = max(0, total_cases[i]   - total_cases[i-1])
        old_new_cases = max(0, total_cases[i-1] - total_cases[i-2])

        alpha += new_cases
        beta  += old_new_cases
        v_alpha.append(alpha)
        v_beta.append(beta)

        Rt_est   = max(0, 1 + infectious_period*np.log(Gamma.mean(     a = alpha, scale = 1/beta)))
        Rt_upper = max(0, 1 + infectious_period*np.log(Gamma.ppf(CI,   a = alpha, scale = 1/beta)))
        Rt_lower = max(0, 1 + infectious_period*np.log(Gamma.ppf(1-CI, a = alpha, scale = 1/beta)))
        Rt_pred.append(Rt_est)
        Rt_CI_upper.append(Rt_upper)
        Rt_CI_lower.append(Rt_lower)

        if (new_cases == 0 or old_new_cases == 0):
            if new_cases == 0:
                logger.debug("new_cases at time %s: 0", i)
            if old_new_cases == 0:
                logger.debug("old_new_cases at time %s: 0", i)
            T_pred.append(0)
            T_CI_upper.append(10) # <- where does this come from?
            T_CI_lower.append(0)
            new_cases_ts.append(0)

        if (new_cases > 0 and old_new_cases > 0):
            new_cases_ts.append(new_cases)

            r, p = alpha, beta/(old_new_cases + beta)
            T_pred.append(nbinom.mean(r, p))
            T_upper = nbinom.ppf(CI,   r, p)
            T_lower = nbinom.ppf(1-CI, r, p)
            T_CI_upper.append(T_upper)
            T_CI_lower.append(T_lower)

            _np = p
            _nr = r 
            anomaly_noted = False
            counter = 0
            while not (T_lower < new_cases < T_upper):
                if not anomaly_noted:
                    anomalies.append(new_cases)
                    anomaly_dates.append(dates[i])
                
                # logger.debug("anomaly identified at time %s: %s < %s < %s, r: %s, p: %s, annealing iteration: %s", i, T_lower, new_cases, T_upper, _nr, _np, counter+1)
                # nnp = 0.95 *_np # <- where does this come from 
                _nr = variance_shift * _nr * ((1-_np)/(1-variance_shift*_np) )
                _np = variance_shift * _np 
                T_upper = nbinom.ppf(CI,   _nr, _np)
                T_lower = nbinom.ppf(1-CI, _nr, _np)
                T_lower, T_upper = sorted((T_lower, T_upper))
                if T_lower == T_upper == 0:
                    T_upper = 1
                    logger.debug("CI collapse, setting T_upper -> 1")
                anomaly_noted = True

                counter += 1
                if counter >= 10000:
                    raise ValueError("Number of iterations exceeded")
            else:
                if anomaly_noted:
                    alpha = _nr # update distribution on R with new parameters that enclose the anomaly 
                    beta = _np/(1-_np) * old_new_cases

                    T_pred[-1] = nbinom.mean(_nr, _np)
                    T_CI_lower[-1] = nbinom.ppf(CI,   _nr, _np)
                    T_CI_upper[-1] = nbinom.ppf(1-CI, _nr, _np)

                    # annealing leaves the RR mean unchanged, but we need to adjust its widened CI
                    Rt_upper = max(0, 1 + infectious_period * np.log(Gamma.ppf(CI    , a = alpha, scale = 1/beta)))
                    Rt_lower = max(0, 1 + infectious_period * np.log(Gamma.ppf(1 - CI, a = alpha, scale = 1/beta)))

                    # replace latest CI time series entries with adjusted CI 
                    Rt_CI_upper[-1] = Rt_upper
                    Rt_CI_lower[-1] = Rt_lower
    return (
        dates[2:], 
        Rt_pred, Rt_CI_upper, Rt_CI_lower, 
        T_pred, T_CI_upper, T_CI_lower, 
        total_cases, new_cases_ts, 
        anomalies, anomaly_dates
    )

def parametric_scheme_mcmc(daily_cases, CI = 0.95, gamma = 0.2, chains = 4, tune = 1000, draws = 1000, **kwargs):
    """ Implements the Bettencourt/Soman parametric scheme via MCMC sampling """
    if isinstance(daily_cases, (pd.DataFrame, pd.Series)):
        case_values = daily_cases.values
    else: 
        case_values = np.array(daily_cases)
    with pm.Model() as mcmc_model:
        # lag new case counts
        dT_lag0 = case_values[1:]
        dT_lag1 = case_values[:-1]
        n = len(dT_lag0)

        dT = pm.Poisson("dT", mu = dT_lag0, shape = (n,))
        bt = pm.Gamma("bt", alpha = dT_lag0.cumsum(), beta = 0.0001 + dT_lag1.cumsum(), shape = (n,))
        Rt = pm.Deterministic("Rt", 1 + pm.math.log(bt)/gamma)
    
        trace = pm.sample(model = mcmc_model, chains = chains, tune = tune, draws = draws, cores = 1, **kwargs)
        return (mcmc_model, trace, pm.summary(trace, hdi_prob = CI))

def branching_random_walk(daily_cases, CI = 0.95, gamma = 0.2, chains = 4, tune = 1000, draws = 1000, **kwargs):
    """ estimate Rt using a random walk for branch parameter, adapted from old Rt.live code """
    if isinstance(daily_cases, (pd.DataFrame, pd.Series)):
        case_values = daily_cases.values
    else: 
        case_values = np.array(daily_cases)
    with pm.Model() as mcmc_model:
        # lag new case counts
        dT_lag0 = case_values[1:]
        dT_lag1 = case_values[:-1]
        n = len(dT_lag0)
        
        # Random walk magnitude
        step_size = pm.HalfNormal('step_size', sigma = 0.03)
        theta_raw_init = pm.Normal('theta_raw_init', 0.1, 0.1)
        theta_raw_steps = pm.Normal('theta_raw_steps', shape = n - 1) * step_size
        theta_raw = tt.concatenate([[theta_raw_init], theta_raw_steps])
        theta = pm.Deterministic('theta', theta_raw.cumsum())

        Rt = pm.Deterministic("Rt", 1 + theta/gamma)
        expected_cases = pm.Poisson('dT', mu = dT_lag1 * pm.math.exp(theta), observed = dT_lag0)
    
        trace = pm.sample(model = mcmc_model, chains = chains, tune = tune, draws = draws, cores = 1, **kwargs)
        return (mcmc_model, trace, pm.summary(trace, hdi_prob = CI))

def linear_projection(dates, R_values, smoothing, period = 7*days):
    """ return 7-day linear projection """
    julian_dates = [_.to_julian_date() for _ in dates[-smoothing//2:None]]
    return OLS(
        R_values[-smoothing//2:None], 
        add_constant(julian_dates)
    )\
    .fit()\
    .predict([1, julian_dates[-1] + period])[0]



from .commons import download_data

""" download the latest policy stringency data from the Oxford tracker """

URL      = "https://raw.githubusercontent.com/OxCGRT/covid-policy-tracker/master/data/"
filename = "OxCGRT_latest.csv"

def download_latest_stringency(dest):
    download_data(dest, filename, base_url = URL)


import datetime
from collections import namedtuple
from pathlib import Path
from typing import Optional, Sequence, Tuple

import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import tikzplotlib
from matplotlib.patheffects import Normal, Stroke
from matplotlib.pyplot import *

from .models import SIR

def normalize_dates(dates):
    try: 
        return [_.to_pydatetime().date() for _ in dates]
    except AttributeError:
        return dates

_ = plt # make mpl package available in epimargin.plots

# default settings
sns.despine()
mpl.rcParams["savefig.dpi"]     = 300
mpl.rcParams["xtick.labelsize"] = "large"
mpl.rcParams["ytick.labelsize"] = "large"
mpl.rcParams["svg.fonttype"]    = "none"

# palettes
## Rt
### core plot
BLK    = "#292f36"
BLK_CI = "#aeb7c2"

### stoplight 
RED = "#D63231"
YLW = "#FD8B5A"
GRN = "#38AE66"

## new cases
OBS_BLK     = BLK
CASE_BLU    = "#335970"
ANOMALY_BLU = "#4092A0"
ANOMALY_RED = "#D63231"
PRED_PURPLE = "#554B68"

## policy simulations 
SIM_PALETTE = ["#437034", "#7D4343", "#43587D", "#7D4370"]

# typography
def rebuild_font_cache():
    import matplotlib.font_manager
    matplotlib.font_manager._rebuild()

def despine(**kwargs):
    pass 

def grid(flag):
    if flag:
        pass
    else:
        pass

# container class for different theme
Aesthetics = namedtuple(
    "Aesthetics", 
    ["title", "label", "note", "ticks", "style", "palette", "accent", "despine", "framealpha", "handlelength"]
)

twitter_settings = Aesthetics(
    title   = {"size": 28, "family": "Overpass", "weight": "bold"},
    label   = {"size": 20, "family": "Overpass", "weight": "regular"},
    note    = {"size": 14, "family": "Overpass", "weight": "regular"},
    ticks   = {"size": 12, "family": "Overpass", "weight": "regular"},
    style   = "white",
    accent  = "dimgrey",
    palette = "bright",
    despine = True,
    framealpha = 0,
    handlelength = 0.5
)

substack_settings = Aesthetics(
    title   = {"size": 28, "family": "SF Pro Display", "weight": "medium"},
    label   = {"size": 20, "family": "SF Pro Display", "weight": "light"},
    note    = {"size": 14, "family": "SF Pro Display", "weight": "light"},
    ticks   = {"size": 12, "family": "Spectral"},
    style   = "white",
    accent  = "#8C8475",
    palette = "bright",
    despine = True,
    framealpha = 0,
    handlelength = 0.5
)

theme = default_settings = Aesthetics(
    title   = {"size": 28, "family": "Helvetica Neue", "weight": "500"},
    label   = {"size": 20, "family": "Helvetica Neue", "weight": "500"},
    note    = {"size": 14, "family": "Helvetica Neue", "weight": "500"},
    ticks   = {"size": 10, "family": "Helvetica Neue"},
    style   = "whitegrid",
    palette = "bright",
    accent  = "dimgrey",
    despine = False,
    framealpha = 1,
    handlelength = 1
)

minimal_settings = Aesthetics(
    title   = {"size": 28, "family": "Helvetica Neue", "weight": "500"},
    label   = {"size": 20, "family": "Helvetica Neue", "weight": "500"},
    note    = {"size": 14, "family": "Helvetica Neue", "weight": "500"},
    ticks   = {"size": 12, "family": "Helvetica Neue"},
    style   = "white",
    palette = "bright",
    accent  = "dimgrey",
    despine = True,
    framealpha = 0,
    handlelength = 0.5
)

plt.rcParams['mathtext.default'] = 'regular'
DATE_FMT = mdates.DateFormatter('%d %b')
bY_FMT   = mdates.DateFormatter('%b %Y')

def set_theme(name):
    global theme
    if   name == "twitter":
        theme = twitter_settings
    elif name == "substack":
        theme = substack_settings
    elif name == "minimal":
        theme = minimal_settings
    else: # default 
        theme = default_settings
    sns.set(style = theme.style, palette = theme.palette, font = theme.ticks["family"])
    mpl.rcParams.update({"font.size": 22})
    if theme.despine:
        plt.rc("axes.spines", top = False, right = False)
    return theme

set_theme("default")


# from https://towardsdatascience.com/beautiful-custom-colormaps-with-matplotlib-5bab3d1f0e72
def hex_to_rgb(value):
    '''
    Converts hex to rgb colours
    value: string of 6 characters representing a hex colour.
    Returns: list length 3 of RGB values'''
    value = value.strip("#") # removes hash symbol if present
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

def rgb_to_dec(value):
    '''
    Converts rgb to decimal colours (i.e. divides each value by 256)
    value: list (length 3) of RGB values
    Returns: list (length 3) of decimal values'''
    return [v/256 for v in value]

def get_continuous_cmap(hex_list, float_list=None):
    ''' creates and returns a color map that can be used in heat map figures.
        If float_list is not provided, colour map graduates linearly between each color in hex_list.
        If float_list is provided, each color in hex_list is mapped to the respective location in float_list. 
        
        Parameters
        ----------
        hex_list: list of hex code strings
        float_list: list of floats between 0 and 1, same length as hex_list. Must start with 0 and end with 1.
        
        Returns
        ----------
        colour map'''
    rgb_list = [rgb_to_dec(hex_to_rgb(i)) for i in hex_list]
    if not float_list:
        float_list = list(np.linspace(0,1,len(rgb_list)))
        
    cdict = dict()
    for num, col in enumerate(['red', 'green', 'blue']):
        col_list = [[float_list[i], rgb_list[i][num], rgb_list[i][num]] for i in range(len(float_list))]
        cdict[col] = col_list
    cmp = mpl.colors.LinearSegmentedColormap('ACRt', segmentdata=cdict, N=256)
    mpl.cm.register_cmap("ACRt", cmp)
    return cmp


# DEFAULT COLOR MAPPING
default_cmap = get_continuous_cmap([GRN, YLW, RED, RED], [0, 0.8, 0.9, 1])
def get_cmap(vmin = 0, vmax = 3, cmap = default_cmap):
    return mpl.cm.ScalarMappable(
        norm = mpl.colors.Normalize(vmin, vmax), 
        cmap = cmap
    )

sm = get_cmap()

def set_tick_size(size: int):
    plt.xticks(fontsize=size)
    plt.yticks(fontsize=size)

# simple wrapper over plt to help chain commands
class PlotDevice():
    def __init__(self, fig: Optional[mpl.figure.Figure] = None):
        self.figure = fig if fig else plt.gcf()
        if theme.despine:
            sns.despine(top = True, right = True)

    def axis_labels(self, x, y, enforce_spacing = True, **kwargs):
        kwargs["fontdict"] = kwargs.get("fontdict", theme.label)
        if enforce_spacing and not x.startswith("\n"):
            x = "\n" + x 
        if enforce_spacing and not y.endswith("\n"):
            y = y + "\n"
        return self.xlabel(x, **kwargs).ylabel(y, **kwargs)

    def xlabel(self, xl: str, **kwargs):
        kwargs["fontdict"] = kwargs.get("fontdict", theme.label)
        plt.xlabel(xl, **kwargs)
        plt.gca().xaxis.label.set_color("dimgray")
        return self 

    def ylabel(self, yl: str, **kwargs):
        kwargs["fontdict"] = kwargs.get("fontdict", theme.label)
        plt.ylabel(yl, **kwargs)
        plt.gca().yaxis.label.set_color("dimgray")
        return self 

    # stack title/subtitle vertically
    def title(self, text: str, **kwargs):
        try:
            kwargs["x"]  = kwargs.get("x", self.figure.get_axes()[0].get_position().bounds[0])
        except IndexError:
            kwargs["x"]  = kwargs.get("x", plt.gca().get_position().bounds[0])
        kwargs["ha"] = kwargs.get("ha", "left")
        kwargs["va"] = kwargs.get("va", "top")
        kwargs["fontsize"]   = kwargs.get("fontsize", theme.title["size"])
        kwargs["fontdict"]   = kwargs.get("fontdict", theme.title)
        kwargs["fontweight"] = kwargs.get("fontweight", theme.title["weight"])
        plt.suptitle(text, **kwargs)
        return self 
    
    def annotate(self, text: str, **kwargs):
        kwargs["fontdict"] = kwargs.get("fontdict", theme.note)
        kwargs["loc"] = kwargs.get("loc", "left")
        plt.title(text, **kwargs)
        return self

    # stack title/subtitle horizontally 
    def l_title(self, text: str, **kwargs):
        kwargs["loc"]        = "left"
        kwargs["ha"]         = kwargs.get("ha", "left")
        kwargs["va"]         = kwargs.get("va", "bottom")
        kwargs["fontsize"]   = kwargs.get("fontsize",   theme.title["size"])
        kwargs["fontdict"]   = kwargs.get("fontdict",   theme.title)
        kwargs["fontweight"] = kwargs.get("fontweight", theme.title["weight"])
        plt.title(text, **kwargs)
        return self 
    
    def r_title(self, text: str, **kwargs):
        kwargs["loc"]      = "right"
        kwargs["ha"]       = kwargs.get("ha", "right")
        kwargs["va"]       = kwargs.get("va", "bottom")
        kwargs["fontdict"] = kwargs.get("fontdict", theme.note)
        kwargs["color"]    = theme.accent
        plt.title(text, **kwargs)
        return self 
    
    def size(self, w, h):
        self.figure.set_size_inches(w, h)
        return self
    
    def legend(self, *args, **kwargs):
        kwargs["framealpha"]   = kwargs.get("framealpha",   theme.framealpha)
        kwargs["handlelength"] = kwargs.get("handlelength", theme.handlelength)
        plt.legend(*args, **kwargs)
        return self
    
    def format_xaxis(self, fmt = DATE_FMT):
        plt.gca().xaxis.set_major_formatter(DATE_FMT)
        plt.gca().xaxis.set_minor_formatter(DATE_FMT)
        return self 

    def save(self, filename: Path, **kwargs):
        if str(filename).endswith("tex"):
            tikzplotlib.save(filename, **kwargs)
            return self 
        kwargs["transparent"] = kwargs.get("transparent", str(filename).endswith("svg"))
        plt.savefig(filename, **kwargs)
        return self 

    def adjust(self, **kwargs):
        plt.subplots_adjust(**kwargs)
        return self 

    def show(self, **kwargs):
        plt.show(**kwargs)
        return self 

def plot_SIRD(model: SIR, layout = (1,  4)) -> PlotDevice:
    """ plot all 4 available curves (S, I, R, D) for a given SIR  model """
    fig, axes = plt.subplots(layout[0], layout[1], sharex = True, sharey = True)
    t = list(range(len(model.Rt)))
    for (ax, model) in zip(axes.flat, model.units):
        s = ax.semilogy(t, model.S, alpha=0.75, label="Susceptibles")
        i = ax.semilogy(t, model.I, alpha=0.75, label="Infectious", )
        d = ax.semilogy(t, model.D, alpha=0.75, label="Deaths",     )
        r = ax.semilogy(t, model.R, alpha=0.75, label="Recovered",  )
        ax.label_outer()
    
    fig.legend([s, i, r, d], labels = ["S", "I", "R", "D"], loc =  "center right", borderaxespad = 0.1)
    return PlotDevice(fig)

def plot_curve(models: Sequence[SIR], labels: Sequence[str], curve: str = "I"):
    """ plot specific epidemic curve """
    fig = plt.figure()
    for (model, label) in zip(models, labels):
        plt.semilogy(model.aggregate(curve), label = label, figure = fig)
    plt.legend() 
    plt.tight_layout()
    return PlotDevice(fig)

def gantt_chart(gantt_data, start_date: Optional[str] = None, show_cbar = True):
    """ create a Gantt chart showing adaptive control status (red/yellow/green) from set of simulations """
    gantt_df = pd.DataFrame(gantt_data, columns = ["district", "day", "beta", "R"])
    gantt_pv = gantt_df.pivot("district", "day", values = ["beta", "R"])
    if start_date:
        start_timestamp = pd.to_datetime(start_date)
        dates = [start_timestamp + datetime.timedelta(days = n) for n in gantt_df.day.unique()]
        xticklabels = [str(xl.day) + " " + xl.month_name()[:3] for xl in dates]
        xlabel = "Date"
    else:
        xticklabels = sorted(gantt_df.day.unique())
        xlabel = "Days Since Beginning of Adaptive Control"
    ax = sns.heatmap(gantt_pv["beta"], linewidths = 2, alpha = 0.8, 
        annot = gantt_pv["R"], annot_kws={"size": 8},
        cmap = ["#38AE66", "#FFF3B4", "#FD8B5A", "#D63231"],
        cbar = show_cbar,
        yticklabels = gantt_df["district"].unique(),
        xticklabels = xticklabels,
        cbar_kws = {
            "ticks":[0.5, 1, 2, 2.5], 
            "label": "Mobility", 
            "format": mpl.ticker.FuncFormatter(lambda x, pos: {0.5:"voluntary", 1:"cautionary", 2:"partial", 2.5:"restricted"}[x]), 
            "orientation": "horizontal", 
            "aspect": 50, 
            "drawedges": True,
            "fraction": 0.05,
            "pad": 0.10, 
            "shrink": 0.5
        }
    )
    plt.tight_layout()
    plt.gcf().subplots_adjust(left=0.10, bottom=0.10)
    plt.xlabel(xlabel, {"size": 20, "family": "Fira Sans", "fontweight": "500"})
    plt.ylabel(None)

    return PlotDevice()

def predictions(date_range, model, color, bounds = [2.5, 97.5], curve = "dT"):
    mdn, min_, max_ = zip(*[np.percentile(_, [50] + bounds) for _ in model.__getattribute__(curve)])
    range_marker   = plt.fill_between(date_range, min_, max_, color = color, alpha = 0.3)
    median_marker, = plt.plot(date_range, mdn, color = color)
    return [(range_marker, median_marker), model.name]

def simulations(
    simulation_results: Sequence[Tuple[SIR]], 
    labels: Sequence[str], 
    historical: Optional[pd.Series] = None, 
    historical_label: str = "Empirical Case Data", 
    curve: str = "dT", 
    smoothing: Optional[np.ndarray] = None, 
    semilog: bool = True) -> PlotDevice:
    """ plot simulation results for new daily cases and optionally show historical trends """

    aggregates = [tuple(model.aggregate(curve) for model in model_set) for model_set in simulation_results]

    policy_outcomes = list(zip(*aggregates))

    num_sims   = len(simulation_results)
    total_time = len(policy_outcomes[0][0])

    ranges = [{"max": [], "min": [], "mdn": [], "avg": []} for _ in range(len(policy_outcomes))]

    for (i, policy) in enumerate(policy_outcomes):
        for t in range(total_time):
            curve_sorted = sorted([curve[t] for curve in policy])
            ranges[i]["min"].append(curve_sorted[0])
            ranges[i]["max"].append(curve_sorted[-1])
            ranges[i]["mdn"].append(curve_sorted[num_sims//2])
            ranges[i]["avg"].append(np.mean(curve_sorted))

    legends = []
    legend_labels  = []
    if historical is not None:
        p, = plt.plot(historical.index, historical, 'k-', alpha = 0.8, zorder = 10)
        t  = [historical.index.max() + datetime.timedelta(days = n) for n in range(total_time)]
        legends.append(p)
        legend_labels.append(historical_label)
    else:
        t = list(range(total_time))

    if smoothing is not None:
        p = plt.plot([pd.Timestamp(t) for t in smoothing[:, 0]], smoothing[:, 1], 'k-', linewidth = 1)
        legends.append(p)
        legend_labels.append("smoothed_data")
        
    for (rng, label, color) in zip(ranges, labels, SIM_PALETTE):
        p, = plt.plot(t, rng["avg"], color = color, linewidth = 2)
        f  = plt.fill_between(t, rng["min"], rng["max"], color = color, alpha = 0.2)
        legends.append((p, f))
        legend_labels.append(label)
    
    plt.gca().xaxis.set_major_formatter(DATE_FMT)
    plt.gca().xaxis.set_minor_formatter(DATE_FMT)
    plt.legend(legends, legend_labels, prop = dict(size = 20), handlelength = theme.handlelength, framealpha = theme.framealpha, loc = "best")

    plt.xlim(left = historical.index[0], right = t[-1])
    if semilog:
        plt.semilogy()
    set_tick_size(14)
    return PlotDevice()

def Rt(dates, Rt_pred, Rt_CI_upper, Rt_CI_lower, CI, ymin = 0.5, ymax = 3, yaxis_colors = True, format_dates = True, critical_threshold = True, legend = True, legend_loc = "best"):
    """ plot Rt and associated confidence  intervals over time """
    CI_marker  = plt.fill_between(dates, Rt_CI_lower, Rt_CI_upper, color = BLK, alpha = 0.3)
    Rt_marker, = plt.plot(dates, Rt_pred, color = BLK, linewidth = 2, zorder = 5, solid_capstyle = "butt")
    if yaxis_colors: 
        plt.plot([dates[0], dates[0]], [2.5, ymax], color = RED, linewidth = 6, alpha = 0.9, solid_capstyle="butt", zorder = 10)
        plt.plot([dates[0], dates[0]], [1,    2.5], color = YLW, linewidth = 6, alpha = 0.9, solid_capstyle="butt", zorder = 10)
        plt.plot([dates[0], dates[0]], [ymin,   1], color = GRN, linewidth = 6, alpha = 0.9, solid_capstyle="butt", zorder = 10)
        plt.plot([dates[0], dates[0]], [ymin, ymax], color = "white", linewidth = 10, alpha = 1, solid_capstyle="butt", zorder = 9)
    if critical_threshold:
        plt.hlines(1, xmin=dates[0], xmax=dates[-1], zorder = 11, color = "black", linestyles = "dotted")
    plt.ylim(ymin, ymax)
    plt.xlim(left=dates[0], right=dates[-1])
    pd = PlotDevice()
    if legend:
        pd.legend_props = dict(framealpha = theme.framealpha, handlelength = theme.handlelength, loc = legend_loc)
        plt.legend([(CI_marker, Rt_marker)], [f"Estimated $R_t$ ({100*CI}% CI)"], **pd.legend_props)
    if format_dates:
        plt.gca().xaxis.set_major_formatter(DATE_FMT)
        plt.gca().xaxis.set_minor_formatter(DATE_FMT)
    set_tick_size(theme.ticks["size"])
    pd.markers = {"Rt" : (CI_marker, Rt_marker)}
    return pd 

def daily_cases(dates, T_pred, T_CI_upper, T_CI_lower, new_cases_ts, anomaly_dates, anomalies, CI, prediction_ts = None): 
    """ plots expected, smoothed cases from simulated annealing training """
    new_cases_dates = dates[-len(new_cases_ts):]
    exp_cases_dates = dates[-len(T_pred):]
    valid_idx   = [i for i in range(len(dates)) if dates[i] not in anomaly_dates] 
    T_CI_lower_rect = [min(l, u) for (l, u) in zip(T_CI_lower, T_CI_upper)]
    T_CI_upper_rect = [max(l, u) for (l, u) in zip(T_CI_lower, T_CI_upper)]
    observed_marker,   = plt.plot([d for d in new_cases_dates if d not in anomaly_dates], [new_cases_ts[i] for i in range(len(new_cases_ts)) if new_cases_dates[i] not in anomaly_dates], color = OBS_BLK, linewidth = 2, zorder = 8)
    anomalies_marker,  = plt.plot(anomaly_dates, anomalies, marker="o", mfc = "none", mec=ANOMALY_RED, lw = 0, zorder = 15)
    expected_marker,   = plt.plot(dates[-len(T_pred):], T_pred, color = CASE_BLU, marker=".", zorder = 10, lw = 0)
    expected_CI_marker = plt.fill_between(dates, T_CI_lower_rect, T_CI_upper_rect, label = f"", facecolor = CASE_BLU, alpha = 0.35)
    legends = [observed_marker, (expected_CI_marker, expected_marker)]
    labels  = ["observed cases (smoothed)", f"expected cases ({100*CI}% CI)"]
    (_, top) = plt.ylim()
    if prediction_ts:
        t_pred = [(dates[-1] + datetime.timedelta(days = i)) for i in range(len(prediction_ts[0][0]))]
        end = t_pred[-1]
        for (predictions, pred_CI_lower, pred_CI_upper, color, label) in prediction_ts:
            predicted_marker,   = plt.plot(t_pred, predictions, color = color, marker = ".", zorder = 9, lw = 0)
            predicted_CI_marker = plt.fill_between(t_pred, pred_CI_lower, pred_CI_upper, facecolor = color, alpha = 0.35)
            (_, top) = plt.ylim()
            plt.vlines(dates[-1], ymin = 0, ymax = top, color = "black", linestyles = "dotted")
            legends += [(predicted_CI_marker, predicted_marker)]
            labels  += [label]
    else: 
        end = dates[-1]
    plt.ylim(bottom = 0, top = top)
    legends += [anomalies_marker]
    labels  += ["anomalies"]
    xlim(left = dates[0], right = end)
    plt.legend(legends, labels, prop = {'size': 14}, framealpha = theme.framealpha, handlelength = theme.handlelength, loc = "best")
    plt.gca().xaxis.set_major_formatter(DATE_FMT)
    plt.gca().xaxis.set_minor_formatter(DATE_FMT)
    set_tick_size(14)
    return PlotDevice()

def choropleth(gdf, label_fn = lambda _: "", col = "Rt", title = "$R_t$", label_kwargs = {}, mappable = sm, fig = None, ax = None):
    """ display choropleth of locations by metric """
    gdf["pt"] = gdf["geometry"].centroid
    if not fig:
        fig, ax = plt.subplots()
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])
    if title:
        ax.set_title(title, loc="left", fontdict = theme.label) 
    gdf.plot(color=[mappable.to_rgba(_) for _ in gdf[col]], ax = ax, edgecolors="black", linewidth=0.5, missing_kwds = {"color": theme.accent, "edgecolor": "white"})
    if label_fn is not None:
        for (_, row) in gdf.iterrows():
            label = label_fn(row)
            value = round(row[col], 2)
            ax.annotate(
                s = f"{label}{value}", 
                xy = list(row["pt"].coords)[0], 
                ha = "center", 
                fontfamily = theme.note["family"], 
                color = "black", **label_kwargs, 
                fontweight = "semibold",
                size = 12)\
                .set_path_effects([Stroke(linewidth = 2, foreground = "white"), Normal()])
    cbar_ax = fig.add_axes([0.90, 0.25, 0.01, 0.5])
    cb = fig.colorbar(mappable = mappable, orientation = "vertical", cax = cbar_ax)
    cbar_ax.set_title("$R_t$", fontdict = theme.note)
    
    return PlotDevice(fig)

def double_choropleth(gdf, label_fn = lambda _: "", Rt_col = "Rt", Rt_proj_col = "Rt_proj", titles = ["Current $R_t$", "Projected $R_t$ (1 Week)"], arrangement = (1, 2), label_kwargs = {}, mappable = sm):
    """ plot two choropleths side-by-side based on multiple metrics """
    gdf["pt"] = gdf["geometry"].centroid
    fig, (ax1, ax2) = plt.subplots(*arrangement)
    for (ax, title, col) in zip((ax1, ax2), titles, (Rt_col, Rt_proj_col)):
        ax.grid(False)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(title, loc="left", fontdict = theme.label) 
        gdf.plot(color=[mappable.to_rgba(_) for _ in gdf[col]], ax = ax, edgecolors="black", linewidth=0.5, missing_kwds = {"color": theme.accent, "edgecolor": "white"})
    if label_fn is not None:
        for (_, row) in gdf.iterrows():
            label = label_fn(row)
            Rt_c, Rt_p = round(row[Rt_col], 2), round(row[Rt_proj_col], 2)
            a1 = ax1.annotate(s=f"{label}{Rt_c}", xy=list(row["pt"].coords)[0], ha = "center", fontfamily = theme.note["family"], color="white", **label_kwargs)
            a2 = ax2.annotate(s=f"{label}{Rt_p}", xy=list(row["pt"].coords)[0], ha = "center", fontfamily = theme.note["family"], color="white", **label_kwargs)
            a1.set_path_effects([Stroke(linewidth = 2, foreground = "black"), Normal()])
            a2.set_path_effects([Stroke(linewidth = 2, foreground = "black"), Normal()])
    cbar_ax = fig.add_axes([0.95, 0.25, 0.01, 0.5])
    cb = fig.colorbar(mappable = mappable, orientation = "vertical", cax = cbar_ax)
    cbar_ax.set_title("$R_t$", fontdict = theme.note)
    
    return PlotDevice(fig)

def double_choropleth_v(*args, **kwargs):
    """ plot two choropleths (one on top of the other) based on multiple metrics """
    kwargs["arrangement"] = (2, 1)
    return double_choropleth(*args, **kwargs)

double_choropleth.horizontal = double_choropleth
double_choropleth.vertical   = double_choropleth_v



import argparse
import logging
import sys
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd

# code readability
days     = 1
weeks    = 7 
years    = 365
percent  = 1/100
annually = 1/years
million  = 1e6

def cwd() -> Path:
    argv0 = sys.argv[0]
    if argv0.endswith("ipython"):
        return Path(".").resolve()
    return Path(argv0).resolve().parent
        
def fmt_params(**kwargs) -> str:
    """  get useful experiment tag from a dictionary of experiment settings """
    return ", ".join(f"{k.replace('_', ' ')}: {v}" for (k, v) in kwargs.items())

def assume_missing_0(df: pd.DataFrame, col: str):
    return df[col] if col in df.columns else 0

def mkdir(p: Path, exist_ok: bool = True) -> Path:
    p.mkdir(exist_ok = exist_ok)
    return p

def setup(**kwargs) -> Tuple[Path]:
    root = cwd()
    if len(sys.argv) > 2:
        parser = argparse.ArgumentParser()
        parser.add_argument("--level", type=str)
        flags = parser.parse_args()
        kwargs["level"] = flags.level
    logging.basicConfig(**kwargs)
    logging.getLogger('flat_table').addFilter(lambda _: 0)
    return (mkdir(root / "data"), mkdir(root / "figs"))

def fillna(array):
    return np.nan_to_num(array, nan = 0, posinf = 0, neginf = 0)

def normalize(array, axis = 0):
    return fillna(array/array.sum(axis = axis)[:, None])



from pathlib import Path
from typing import Optional, Sequence

import numpy as np
import pandas as pd

from ..utils import assume_missing_0

""" tools to extract time series for COVID19India.org data """

# states created after the 2001 census
new_states = set("Telangana")

# states renamed in 2011
renamed_states = { 
    "Orissa"      : "Odisha",
    "Pondicherry" : "Puducherry"
}

columns_v1 = v1 = [
    "patient number",
    "state patient number",
    "date announced",
    "age bracket",
    "gender",
    "detected city",
    "detected district",
    "detected state",
    "current status",
    "notes",
    "contracted from which patient (suspected)",
    "nationality",
    "type of transmission",
    "status change date",
    "source_1",
    "source_2",
    "source_3",
    "backup note"
]

columns_v2 = v2 = [
    'patient number',
    'state patient number',
    'date announced',
    'estimated onset date',
    'age bracket',
    'gender',
    'detected city',
    'detected district',
    'detected state',
    'state code',
    'current status',
    'notes',
    'contracted from which patient (suspected)',
    'nationality',
    'type of transmission',
    'status change date',
    'source_1',
    'source_2',
    'source_3',
    'backup notes'
]

drop_cols = {
    "age bracket",
    "gender",
    "detected city",
    # "detected district",
    "notes",
    "contracted from which patient (suspected)",
    "nationality",
    "source_1",
    "source_2",
    "source_3",
    "backup note",
    "backup notes",
    "type of transmission"
}

columns_v3 = v3 = [
    'Patient Number',
    'State Patient Number',
    'Date Announced',
    'Estimated Onset Date',
    'Age Bracket',
    'Gender',
    'Detected City',
    'Detected District',
    'Detected State',
    'State code',
    'Current Status',
    'Notes',
    'Contracted from which Patient (Suspected)',
    'Nationality',
    'Type of transmission',
    'Status Change Date',
    'Source_1',
    'Source_2',
    'Source_3',
    'Backup Notes',
    'Num Cases'
]

drop_cols_v3 = {
    "Age Bracket",
    "Gender",
    "Detected City",
    "Notes",
    'Contracted from which Patient (Suspected)', 
    'Nationality',
    "Source_1",
    "Source_2",
    "Source_3",
    "Backup Notes",
    "State Patient Number",
    "State code",
    "Estimated Onset Date",
    "Type of transmission"
}

columns_v4 = v4 = [
    'Entry_ID', 
    'State Patient Number', 
    'Date Announced', 
    'Age Bracket',
    'Gender', 
    'Detected City', 
    'Detected District', 
    'Detected State',
    'State code', 
    'Num Cases', 
    'Current Status',
    'Contracted from which Patient (Suspected)', 
    'Notes', 
    'Source_1',
    'Source_2', 
    'Source_3', 
    'Nationality', 
    'Type of transmission',
    'Status Change Date', 
    'Patient Number'
]

drop_cols_v4 = {
    "Entry_ID",
    'Age Bracket',
    'Gender', 
    'Detected City',
    'State code',
    'Contracted from which Patient (Suspected)',
    'Notes', 
    'Source_1',
    'Source_2', 
    'Source_3', 
    'Nationality', 
    'Type of transmission',
    "State Patient Number"
}

column_ordering_v4  = [
    'patient_number',
    'date_announced',
    'detected_district',
    'detected_state',
    'current_status',
    'status_change_date',
    'num_cases'
 ]

district_2011_replacements = {
    'Maharashtra' : {
        'Mumbai Suburban' : 'Mumbai'}
 }

state_code_lookup = {
    'AN'  : 'Andaman & Nicobar Islands',
    'AP'  : 'Andhra Pradesh',
    'AR'  : 'Arunachal Pradesh',
    'AS'  : 'Assam',
    'BR'  : 'Bihar',
    'CH'  : 'Chandigarh',
    'CT'  : 'Chhattisgarh',
    'DD'  : 'Daman & Diu', 
    'DNDD': 'Dadra & Nagar Haveli and Daman & Diu',
    'DL'  : 'Delhi',
    'DN'  : 'Dadra & Nagar Haveli',
    'GA'  : 'Goa',
    'GJ'  : 'Gujarat',
    'HP'  : 'Himachal Pradesh',
    'HR'  : 'Haryana',
    'JH'  : 'Jharkhand',
    'JK'  : 'Jammu & Kashmir',
    'KA'  : 'Karnataka',
    'KL'  : 'Kerala',
    'LA'  : 'Ladakh',
    'LD'  : 'Lakshadweep',
    'MH'  : 'Maharashtra',
    'ML'  : 'Meghalaya',
    'MN'  : 'Manipur',
    'MP'  : 'Madhya Pradesh',
    'MZ'  : 'Mizoram',
    'NL'  : 'Nagaland',
    'OR'  : 'Odisha',
    'PB'  : 'Punjab',
    'PY'  : 'Puducherry',
    'RJ'  : 'Rajasthan',
    'SK'  : 'Sikkim',
    'TG'  : 'Telangana',
    'TN'  : 'Tamil Nadu',
    'TR'  : 'Tripura',
    'TT'  : 'India',
    'UN'  : 'State Unassigned',
    'UP'  : 'Uttar Pradesh',
    'UT'  : 'Uttarakhand',
    'WB'  : 'West Bengal',
}

state_name_lookup = { 
    'Andaman & Nicobar Islands'  : 'AN', 
    'Andaman And Nicobar Islands': 'AN', 
    'Andaman and Nicobar Islands': 'AN', 
    'Andhra Pradesh'             : 'AP', 
    'Arunachal Pradesh'          : 'AR', 
    'Assam'                      : 'AS', 
    'Bihar'                      : 'BR', 
    'Chandigarh'                 : 'CH', 
    'Chhattisgarh'               : 'CT', 
    'Daman & Diu'                : 'DD', 
    'Daman And Diu'              : 'DD', 
    'Daman and Diu'              : 'DD', 
    'Delhi'                      : 'DL', 
    'Dadra & Nagar Haveli'       : 'DN', 
    'Dadra And Nagar Haveli'     : 'DN', 
    'Dadra and Nagar Haveli'     : 'DN', 
    'Goa'                        : 'GA', 
    'Gujarat'                    : 'GJ', 
    'Himachal Pradesh'           : 'HP', 
    'Haryana'                    : 'HR', 
    'Jharkhand'                  : 'JH', 
    'Jammu & Kashmir'            : 'JK', 
    'Jammu and Kashmir'          : 'JK', 
    'Jammu And Kashmir'          : 'JK', 
    'Karnataka'                  : 'KA', 
    'Kerala'                     : 'KL', 
    'Ladakh'                     : 'LA', 
    'Lakshadweep'                : 'LD', 
    'Maharashtra'                : 'MH', 
    'Meghalaya'                  : 'ML', 
    'Manipur'                    : 'MN', 
    'Madhya Pradesh'             : 'MP', 
    'Mizoram'                    : 'MZ', 
    'Nagaland'                   : 'NL', 
    'Odisha'                     : 'OR', 
    'Punjab'                     : 'PB', 
    'Puducherry'                 : 'PY', 
    'Rajasthan'                  : 'RJ', 
    'Sikkim'                     : 'SK', 
    'Telangana'                  : 'TG', 
    'Tamil Nadu'                 : 'TN', 
    'Tripura'                    : 'TR', 
    'India'                      : 'TT', 
    'State Unassigned'           : 'UN', 
    'Uttar Pradesh'              : 'UP', 
    'Uttarakhand'                : 'UT',
    'West Bengal'                : 'WB',
    'Dadra And Nagar Haveli And Daman And Diu' : "DNDD"
}


def data_path(i: int):
    return f"raw_data{i}.csv"

def standardize_column_headers(df: pd.DataFrame):
    df.columns = df.columns.str.lower().str.strip().str.replace(" ","_").str.replace('[^a-zA-Z0-9_]', '')

# load data until April 26
def load_data_v3(path: Path, drop = drop_cols_v3):
    cases = pd.read_csv(path, 
        usecols     = set(columns_v3) - drop,
        dayfirst    = True, # source data does not have consistent date format so cannot rely on inference
        parse_dates = ["Date Announced", "Status Change Date"])
    standardize_column_headers(cases)
    return cases

# load data for April 27 - May 09  
def load_data_v4(path: Path, drop = drop_cols_v3):
    cases = pd.read_csv(path, 
        usecols     = set(columns_v4) - drop,
        dayfirst    = True, # source data does not have consistent date format so cannot rely on inference
        parse_dates = ["Date Announced", "Status Change Date"])
    standardize_column_headers(cases)
    return cases[column_ordering_v4]

def add_time_col(grp_df):
    grp_df['time'] = (grp_df["date"] - grp_df["date"].min()).dt.days
    return grp_df

# calculate daily totals and growth rate
def get_time_series(df: pd.DataFrame, group_col: Optional[Sequence[str]] = None, drop_negatives = True) -> pd.DataFrame:
    if group_col:
        group_cols = (group_col if isinstance(group_col, list) else [group_col]) + ["status_change_date", "current_status"]
    else: 
        group_cols = ["status_change_date", "current_status"]
    if drop_negatives:
        df = df[df["num_cases"] >= 0]
    totals = df.groupby(group_cols)["num_cases"].agg(lambda counts: np.sum(np.abs(counts)))
    if len(totals) == 0:
        return pd.DataFrame()
    totals = totals.unstack().fillna(0)[['Deceased','Hospitalized','Recovered']]
    totals["date"] = totals.index.get_level_values("status_change_date")
    totals = totals.groupby(level=0).apply(add_time_col) if group_col else add_time_col(totals)
    totals["delta"] = assume_missing_0(totals, "Hospitalized") - assume_missing_0(totals, "Recovered") - assume_missing_0(totals, "Deceased")
    totals["logdelta"] = np.ma.log(totals["delta"].values).filled(0)
    return totals

def load_all_data(v3_paths: Sequence[Path], v4_paths: Sequence[Path]) -> pd.DataFrame:
    cases_v3 = [load_data_v3(path) for path in v3_paths]
    cases_v4 = [load_data_v4(path) for path in v4_paths]
    all_cases = pd.concat(cases_v3 + cases_v4)
    all_cases["status_change_date"] = all_cases["status_change_date"].fillna(all_cases["date_announced"])
    all_cases["detected_state"]     = all_cases["detected_state"].str.strip().str.title()
    all_cases["detected_district"]  = all_cases["detected_district"].str.strip().str.title()  
    return all_cases.dropna(subset  = ["detected_state"])

# assuming analysis for data structure from COVID19-India saved as resaved, properly-quoted file (v1 and v2)
def load_data(datapath: Path, reduced: bool = False, schema: Optional[Sequence[str]] = None) -> pd.DataFrame: 
    if not schema:
        schema = columns_v1
    df = pd.read_csv(datapath, 
        skiprows    = 1, # supply fixed header in order to deal with Google Sheets export issues 
        names       = schema, 
        usecols     = (lambda _: _ not in drop_cols) if reduced else None,
        dayfirst    = True, # source data does not have consistent date format so cannot rely on inference
        parse_dates = ["Date Announced", "Status Change Date"])
    standardize_column_headers(df)
    return df

# replace covid api detected_district names with 2011 district name
def replace_district_names(df_state: pd.DataFrame, state_district_maps: pd.DataFrame) -> pd.DataFrame:
    state_district_maps = state_district_maps[['district_covid_api', 'district_2011']].set_index('district_covid_api')
    district_map_dict = state_district_maps.to_dict()['district_2011']
    df_state['detected_district'].replace(district_map_dict, inplace=True)
    return df_state

def load_statewise_data(statewise_data_path: Path, drop_unassigned: bool = True) -> pd.DataFrame:
    df_raw = pd.read_csv(statewise_data_path, parse_dates = ["Date"])
    df_raw.rename(columns=state_code_lookup, inplace=True)
    df = pd.DataFrame(df_raw.set_index(["Date","Status"]).unstack().unstack()).reset_index()
    df.columns = ["state", "current_status", "status_change_date", "num_cases"]
    df.replace("Confirmed", "Hospitalized", inplace=True)
    # drop negative cases and cases with no state assigned
    df_assigned = df[(~df["state"].isin(["Date_YMD", "State Unassigned", "TT"] if drop_unassigned else ["Date_YMD", "TT"]))]
    return df_assigned[(df_assigned["num_cases"] >= 0)]



from setuptools import setup, find_packages

setup(
    name = 'epimargin',
    version = '0.5.0',
    packages = find_packages(),
    author = "COVID International Working Group",
    description = "Toolkit for estimating epidemiological metrics and evaluating public health and economic policies.",
    url = "https://github.com/COVID-IWG/epimargin",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.6',
    install_requires = [
        "numpy",
        "pandas",
        "matplotlib",
        "arviz",
        "pymc3",
        "statsmodels",
        "flat-table",
        "requests",
        "seaborn",
        "tikzplotlib",
        "geopandas"
    ]
)



from typing import Optional, Sequence

import numpy as np
from scipy.signal import convolve, filtfilt, iirnotch
from statsmodels.nonparametric.smoothers_lowess import lowess as sm_lowess

# supported kernels for convolution smoothing
kernels = { 
    "hanning"  : np.hanning,
    "hamming"  : np.hamming,
    "bartlett" : np.bartlett,
    "blackman" : np.blackman,
    "uniform"  : np.ones
}

def notched_smoothing(window: int = 7):
    """ Removes weekly and twice-weekly periodicity before convolving a time-reversed padded signal with a uniform moving average window"""
    fs, f0, Q = 1, 1/7, 1
    b1, a1 = iirnotch(f0, Q, fs)
    b2, a2 = iirnotch(2*f0, 2*Q, fs)
    # Frequency response
    b = convolve(b1, b2)
    a = convolve(a1, a2)
    kernel = np.ones(window)/window
    def smooth(data: Sequence[float]):
        notched = filtfilt(b, a, data)
        return convolve(np.concatenate([notched, notched[:-window-1:-1]]), kernel, mode="same")[:-window]
    return smooth

def notch_filter():
    """ implements a notch filter with notches at 1/week and 2/week; assuming input signal sampling rate is 1/week"""
    fs, f0, Q = 1, 1/7, 1
    b1, a1 = iirnotch(f0, Q, fs)
    b2, a2 = iirnotch(2*f0, 2*Q, fs)
    # Frequency response
    b = convolve(b1, b2)
    a = convolve(a1, a2)
    def filter_(data: Sequence[float]):
        return filtfilt(b, a, data)
    return filter_

def convolution(key: str = "hamming", window: int = 7):
    """ entry point for all convolution operations """
    kernel = kernels[key](window)
    def smooth(data: Sequence[float]):
        # pad the data with time reversal windows of signal at ends since all kernels here are apodizing 
        padded = np.r_[data[window-1:0:-1], data, data[-2:-window-1:-1]]
        return np.convolve(kernel/kernel.sum(), padded, mode="valid")[:-window+1]
    return smooth

def box_filter_local(window: int = 5, local_smoothing: Optional[int] = 3):
    """ implement a box filter smoother with additional LOWESS-like smoothing for data points at the end of the timeseries"""
    def smooth(data: Sequence[float]):
        smoothed = np.convolve(data, np.ones(window)/window, mode='same')
        if local_smoothing and len(data) > (local_smoothing + 1):
            for i in range(local_smoothing-1, 0, -1):
                smoothed[-i] = np.mean(data[-i-local_smoothing+1: -i+1 if i > 1 else None])
        return smoothed
    return smooth 

def lowess(**kwargs):
    """ wrapper over statsmodels lowess implementation to return a callable """
    return lambda data: sm_lowess(data, list(range(len(data))), **kwargs)



from pathlib import Path
from typing import Dict, Iterator, Optional, Sequence, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
from scipy.spatial import distance_matrix
from scipy.stats import poisson, binom
from .utils import normalize, fillna

"""
Stochastic epidemiological models for forward simulation.
"""

class SIR():
    """ stochastic SIR compartmental model with external introductions """
    def __init__(self, 
        name:                str,           # name of unit
        population:          int,           # unit population
        dT0:        Optional[int]  = None,  # last change in cases, None -> Poisson random intro 
        Rt0:                 float = 1.9,   # initial reproductive rate,
        I0:                  int   = 0,     # initial infected
        R0:                  int   = 0,     # initial recovered
        D0:                  int   = 0,     # initial dead
        S0:         Optional[int]  = None,  # initial susceptibles, calculated from N, I, R, D if not provided
        infectious_period:   int   = 5,     # how long disease is communicable in days 
        introduction_rate:   float = 5.0,   # parameter for new community transmissions (lambda) 
        mortality:           float = 0.02,  # I -> D transition probability 
        mobility:            float = 0,     # percentage of total population migrating out at each timestep 
        upper_CI:            float = 0.0,   # initial upper confidence interval for new case counts
        lower_CI:            float = 0.0,   # initial lower confidence interval for new case counts
        CI:                  float = 0.95,  # confidence interval
        random_seed:         int   = 0      # random seed 
        ):
        
        # save params 
        self.name  = name 
        self.pop0  = population
        self.gamma = 1.0/infectious_period
        self.ll    = introduction_rate
        self.m     = mortality
        self.mu    = mobility
        self.Rt0   = Rt0
        self.CI    = CI 

        # state and delta vectors 
        if dT0 is None:
            dT0 = np.random.poisson(self.ll) # initial number of new cases 
        self.dT = [dT0] # case change rate, initialized with the first introduction, if any
        self.Rt = [Rt0]
        self.b  = [np.exp(self.gamma * (Rt0 - 1.0))]
        self.S  = [S0 if S0 is not None else population - R0 - D0 - I0]
        self.I  = [I0] 
        self.R  = [R0]
        self.D  = [D0]
        self.dR = [0]
        self.dD = [0]
        self.N  = [population - D0] # total population = S + I + R 
        self.beta = [Rt0 * self.gamma] # initial contact rate 
        self.total_cases = [I0] # total cases 
        self.upper_CI = [upper_CI]
        self.lower_CI = [lower_CI]

        np.random.seed(random_seed)

    # period 1: inter-state migratory transmission
    def migration_step(self) -> int:
        # note: update state *in place* since we consider it the same time period 
        outflux = np.random.poisson(self.mu * self.I[-1])
        new_I = self.I[-1] - outflux
        if new_I < 0: new_I = 0
        self.I[-1]  = new_I
        self.N[-1] -= outflux
        return outflux

    # period 2: intra-state community transmission
    def forward_epi_step(self, dB: int = 0): 
        # get previous state 
        S, I, R, D, N = (vector[-1] for vector in (self.S, self.I, self.R, self.D, self.N))

        # update state 
        Rt = self.Rt0 * float(S)/float(N)
        b  = np.exp(self.gamma * (Rt - 1))

        rate_T    = max(0, self.b[-1] * self.dT[-1] + (1 - self.b[-1] + self.gamma * self.b[-1] * self.Rt[-1])*dB)
        num_cases = poisson.rvs(rate_T)
        self.upper_CI.append(poisson.ppf(self.CI,     rate_T))
        self.lower_CI.append(poisson.ppf(1 - self.CI, rate_T))

        I += num_cases
        S -= num_cases

        rate_D    = self.m * self.gamma * I
        num_dead  = poisson.rvs(rate_D)
        D        += num_dead

        rate_R    = (1 - self.m) * self.gamma * I 
        num_recov = poisson.rvs(rate_R)
        R        += num_recov

        I -= (num_dead + num_recov)
        
        if S < 0: S = 0
        if I < 0: I = 0
        if D < 0: D = 0

        N = S + I + R
        beta = (num_cases * N)/(b * S * I)

        # update state vectors 
        self.Rt.append(Rt)
        self.b.append(b)
        self.S.append(S)
        self.I.append(I)
        self.R.append(R)
        self.D.append(D)
        self.dR.append(num_recov)
        self.dD.append(num_dead)
        self.N.append(N)
        self.beta.append(beta)
        self.dT.append(num_cases)
        self.total_cases.append(I + R + D)
    
    # parallel poisson draws for infection
    def parallel_forward_epi_step(self, dB: int = 0, num_sims = 10000): 
        # get previous state 
        S, I, R, D, N = (vector[-1].copy() for vector in (self.S, self.I, self.R, self.D, self.N))

        # update state 
        Rt = self.Rt0 * S/N
        b  = np.exp(self.gamma * (Rt - 1))

        rate_T    = (self.b[-1] * self.dT[-1]).clip(0)
        num_cases = poisson.rvs(rate_T, size = num_sims)
        self.upper_CI.append(poisson.ppf(self.CI,     rate_T))
        self.lower_CI.append(poisson.ppf(1 - self.CI, rate_T))

        I += num_cases
        S -= num_cases

        rate_D    = self.m * self.gamma * I
        num_dead  = poisson.rvs(rate_D, size = num_sims)
        D        += num_dead

        rate_R    = (1 - self.m) * self.gamma * I 
        num_recov = poisson.rvs(rate_R, size = num_sims)
        R        += num_recov

        I -= (num_dead + num_recov)

        S = S.clip(0)
        I = I.clip(0)
        D = D.clip(0)

        N = S + I + R
        beta = (num_cases * N)/(b * S * I)

        # update state vectors 
        self.Rt.append(Rt)
        self.b.append(b)
        self.S.append(S)
        self.I.append(I)
        self.R.append(R)
        self.D.append(D)
        self.dR.append(num_recov)
        self.dD.append(num_dead)
        self.N.append(N)
        self.beta.append(beta)
        self.dT.append(num_cases)
        self.total_cases.append(I + R + D)

    # parallel binomial draws for infection
    def parallel_forward_binom_step(self, dB: int = 0, num_sims = 10000): 
        # get previous state 
        S, I, R, D, N = (vector[-1].copy() for vector in (self.S, self.I, self.R, self.D, self.N))

        # update state 
        Rt = self.Rt0 * S/N
        p = self.gamma * Rt * I/N

        num_cases = binom.rvs(n = S.astype(int), p = p, size = num_sims)
        self.upper_CI.append(binom.ppf(self.CI,     n = S.astype(int), p = p))
        self.lower_CI.append(binom.ppf(1 - self.CI, n = S.astype(int), p = p))

        I += num_cases
        S -= num_cases

        rate_D    = self.m * self.gamma * I
        num_dead  = poisson.rvs(rate_D, size = num_sims)
        D        += num_dead

        rate_R    = (1 - self.m) * self.gamma * I 
        num_recov = poisson.rvs(rate_R, size = num_sims)
        R        += num_recov

        I -= (num_dead + num_recov)

        S = S.clip(0)
        I = I.clip(0)
        D = D.clip(0)

        N = S + I + R
        # beta = (num_cases * N)/(b * S * I)

        # update state vectors 
        self.Rt.append(Rt)
        # self.b.append(b)
        self.S.append(S)
        self.I.append(I)
        self.R.append(R)
        self.D.append(D)
        self.N.append(N)
        # self.beta.append(beta)
        self.dT.append(num_cases)
        self.total_cases.append(I + R + D)

    def run(self, days: int):
        for _ in range(days):
            self.forward_epi_step()
        return self

    def __repr__(self) -> str:
        return f"[{self.name}]"

class Age_SIRVD(SIR):
    """ age-structured compartmental model with a vaccinated class for each age bin """
    def __init__(self,
        name:                str,           # name of unit
        population:          int,           # unit population
        dT0:        Optional[int]  = None,  # last change in cases, None -> Poisson random intro 
        Rt0:                 float = 1.9,   # initial reproductive rate,
        S0:                  int   = 0,     # initial susceptible
        I0:                  int   = 0,     # initial infected
        R0:                  int   = 0,     # initial recovered
        D0:                  int   = 0,     # initial dead
        infectious_period:   int   = 5,     # how long disease is communicable in days 
        introduction_rate:   float = 5.0,   # parameter for new community transmissions (lambda) 
        mortality:           float = 0.02,  # I -> D transition probability 
        mobility:            float = 0,     # percentage of total population migrating out at each timestep 
        upper_CI:            float = 0.0,   # initial upper confidence interval for new case counts
        lower_CI:            float = 0.0,   # initial lower confidence interval for new case counts
        CI:                  float = 0.95,  # confidence interval
        num_age_bins:        int   = 7,     # number of age bins
        phi:                 float = 0.25,  # proportion of population vaccinated annually 
        ve:                  float = 0.7,   # vaccine effectiveness
        random_seed:         int   = 0      # random seed,  
    ):
        super().__init__(name, population, dT0=dT0, Rt0=Rt0, I0=I0, R0=R0, D0=D0, infectious_period=infectious_period, introduction_rate=introduction_rate, mortality=mortality, mobility=mobility, upper_CI=upper_CI, lower_CI=lower_CI, CI=CI, random_seed=random_seed)
        self.N = [S0 + I0 + R0]
        shape = (sims, bins) = S0.shape
        
        self.num_age_bins = num_age_bins
        self.phi  = phi
        self.ve   = ve
        
        self.S    = [S0]

        self.S_vm = [np.zeros(shape)]
        self.S_vn = [np.zeros(shape)]
        
        self.I_vn = [np.zeros(shape)]
        
        self.R_vm = [np.zeros(shape)]
        self.R_vn = [np.zeros(shape)]
        
        self.D_vn = [np.zeros(shape)]

        self.N_vn = [np.zeros(shape)] # number vaccinated, ineffective 
        self.N_vm = [np.zeros(shape)] # number vaccinated, immune 

        self.N_v  = [np.zeros(shape)] # total vaccinated
        self.N_nv = [np.zeros(shape)] # total unvaccinated
        self.pi   = [np.zeros(shape)]
        self.q1   = [np.zeros(shape)]
        self.q0   = [np.zeros(shape)]

        self.dT_total = [np.zeros(sims)]
        self.dD_total = [np.zeros(sims)]
        self.dV = []

        self.rng = np.random.default_rng(random_seed)

    def parallel_forward_epi_step(self, dV: Optional[np.array], num_sims = 10000): 
        """
            in the SIR and NetworkedSIR, the dB is the reservoir introductions; 
            here, dV is a (self.age_bins, num_sims)-sized array of vaccination doses (administered)
        """
        # get previous state 
        S, S_vm, S_vn, I, I_vn, R, R_vm, R_vn, D, D_vn, N, N_vn, N_vm = (_[-1].copy() for _ in 
            (self.S, self.S_vm, self.S_vn, self.I, self.I_vn, self.R, self.R_vm, self.R_vn, self.D, self.D_vn, self.N, self.N_vn, self.N_vm))

        # vaccination occurs here
        dS_vm = (fillna(self.S[-1]/self.N[-1]) * (    self.ve) * dV)
        dS_vn = (fillna(self.S[-1]/self.N[-1]) * (1 - self.ve) * dV)

        dI_vn =  fillna(self.I[-1]/self.N[-1]) * dV
        dR_vm =  fillna(self.R[-1]/self.N[-1]) * dV

        S_vm  = (S_vm + dS_vm).clip(0)
        S_vn  = (S_vn + dS_vn).clip(0)
        S     = (S - (dS_vn + dS_vm)).clip(0)

        I_vn  = (I_vn + dI_vn).clip(0)
        I     = (I    - dI_vn).clip(0)

        R_vm  = (R_vm + dR_vm).clip(0)
        R     = (R    - dR_vm).clip(0)

        S_ratios = normalize(S + S_vn, axis = 1)

        # core epi update with additional bins (infection, death, recovery)
        Rt = self.Rt0 * (S + S_vn).sum(axis = 1)/(N + S_vn + S_vm + I_vn + R_vn + R_vm).sum(axis = 1)
        b  = np.exp(self.gamma * (Rt - 1))

        lambda_T = (self.b[-1] * self.dT[-1]).clip(0)
        dT = self.rng.poisson(lambda_T)
        self.upper_CI.append(poisson.ppf(    self.CI, lambda_T))
        self.lower_CI.append(poisson.ppf(1 - self.CI, lambda_T))

        dS    = fillna(S   /(S+S_vn)) * (S_ratios * dT[:, None])
        dS_vn = fillna(S_vn/(S+S_vn)) * (S_ratios * dT[:, None])

        S    = (S    - dS).clip(0)
        S_vn = (S_vn - dS_vn).clip(0)

        dD    = self.rng.poisson(   self.m  * self.gamma * I   , size = (num_sims, self.num_age_bins))
        dD_vn = self.rng.poisson(   self.m  * self.gamma * I_vn, size = (num_sims, self.num_age_bins))
        dR    = self.rng.poisson((1-self.m) * self.gamma * I   , size = (num_sims, self.num_age_bins))
        dR_vn = self.rng.poisson((1-self.m) * self.gamma * I_vn, size = (num_sims, self.num_age_bins))
        
        dI    = (dS    - (dD    + dR))
        dI_vn = (dS_vn - (dD_vn + dR_vn))

        D    = (D    + dD).clip(0)
        D_vn = (D_vn + dD_vn).clip(0)
        
        R    = (R    + dR).clip(0)
        R_vn = (R_vn + dR_vn).clip(0)
        
        I    = (I    + dI).clip(0)
        I_vn = (I_vn + dI_vn).clip(0)

        N    = S    + I    + R
        N_vn = S_vn + I_vn + R_vn
        N_vm = S_vm +        R_vm 

        # beta = dT[:, None] * N/(b * (S + S_vn) * (I + I_vn))

        # calculate vax policy evaluation metrics 
        N_v  = np.clip((S_vm + S_vn + I_vn + D_vn + R_vn + R_vm), a_min = 0, a_max = self.N[0])
        N_nv = self.N[0] - N_v
        pi   = N_v/self.N[0]

        q1 = np.nan_to_num(1 - (D_vn - self.D_vn[0])/N_v , nan = 0, neginf = 1).clip(0, 1)
        q0 = np.nan_to_num(1 - (D    - self.D   [0])/N_nv, nan = 0, neginf = 1).clip(0, 1)

        # update state vectors 
        self.Rt.append(Rt)
        self.b.append(b)
        self.S.append(S)
        self.S_vm.append(S_vm)
        self.S_vn.append(S_vn)
        self.I.append(I)
        self.I_vn.append(I_vn)
        self.R.append(R)
        self.R_vm.append(R_vm)
        self.R_vn.append(R_vn)
        self.D.append(D)
        self.D_vn.append(D_vn)
        self.dR.append(dR)
        self.dD.append(dD)
        self.N.append(N)
        self.N_vn.append(N_vn)
        self.N_vm.append(N_vm)
        # self.beta.append(beta)
        self.dT.append(dT)
        self.dT_total.append(dT)
        self.dD_total.append((dD + dD_vn).sum(axis = 1))
        self.total_cases.append(I + R + D)
        
        self.N_v.append(N_v)
        self.N_nv.append(N_nv)
        self.pi.append(pi)
        self.q1.append(q1)
        self.q0.append(q0)

        self.dV.append(dV)

class NetworkedSIR():
    """ composition of SIR models implementing cross-geography interactions """
    def __init__(self, units: Sequence[SIR], default_migrations: Optional[np.matrix] = None, random_seed : Optional[int] = None):
        self.units      = units
        self.migrations = default_migrations
        self.names      = {unit.name: unit for unit in units}
        if random_seed is not None:
            np.random.seed(random_seed)

    def __len__(self) -> int:
        return len(self.units)

    def tick(self, migrations: np.matrix):
        # run migration step 
        outflux       = [unit.migration_step() for unit in self.units]
        transmissions = [flux * migrations[i, :].sum() for (i, flux) in enumerate(outflux)]
        
        # now run forward epidemiological model 
        for (unit, tmx) in zip(self.units, transmissions):
            unit.forward_epi_step(tmx)

    def run(self, days: int, migrations: Optional[np.matrix] = None):
        if migrations is None:
            migrations = self.migrations
        for _ in range(days):
            self.tick(migrations)
        return self 

    def __iter__(self) -> Iterator[SIR]:
        return iter(self.units)

    # index units
    def __getitem__(self, idx: Union[str, int]) -> SIR:
        if isinstance(idx, int):
            return self.units[idx]
        return self.names[idx]

    def set_parameters(self, **kwargs):
        for (attr, val) in kwargs.items():
            if callable(val):
                if val.__code__.co_argcount == 1:
                    for unit in self.units:
                        unit.__setattr__(attr, val(unit))
                else: 
                    for (i, unit) in enumerate(self.units):
                        unit.__setattr__(attr, val(i, unit))
            elif isinstance(val, dict):
                for unit in self.units:
                    unit.__setattr__(attr, val[unit.name])
            else: 
                for unit in self.units:
                    unit.__setattr__(attr, val)
        return self 

    def aggregate(self, curves: Union[Sequence[str], str] = ["Rt", "b", "S", "I", "R", "D", "P", "beta"]) -> Union[Dict[str, Sequence[float]], Sequence[float]]:
        if isinstance(curves, str):
            single_curve = curves
            curves = [curves]
        else: 
            single_curve = False
        aggs = { 
            curve: list(map(sum, zip(*(unit.__getattribute__(curve) for unit in self.units))))
            for curve in curves
        }

        if single_curve:
            return aggs[single_curve]
        return aggs

class SEIR():
    """ stochastic SEIR model without external introductions """
    def __init__(self, 
        name:                str,           # name of unit
        population:          int,           # unit population
        dT0:        Optional[int]  = None,  # last change in cases, None -> Poisson random intro 
        Rt0:                 float = 1.9,   # initial reproductive rate,
        E0:                  int   = 0,     # initial exposed
        I0:                  int   = 0,     # initial infected
        R0:                  int   = 0,     # initial recovered
        D0:                  int   = 0,     # initial dead
        infectious_period:   int   = 5,     # how long disease is communicable in days 
        incubation_period:   int   = 5,     # how long the diseas takes to incubate
        introduction_rate:   float = 5.0,   # parameter for new community transmissions (lambda) 
        mortality:           float = 0.02,  # I -> D transition probability 
        mobility:            float = 0,     # percentage of total population migrating out at each timestep 
        upper_CI:            float = 0.0,   # initial upper confidence interval for new case counts
        lower_CI:            float = 0.0,   # initial lower confidence interval for new case counts
        CI:                  float = 0.95,  # confidence interval
        random_seed:         int   = 0      # random seed 
        ):
        
        # save params 
        self.name  = name 
        self.pop0  = population
        self.gamma = 1.0/infectious_period
        self.sigma = 1.0/incubation_period
        self.ll    = introduction_rate
        self.m     = mortality
        self.mu    = mobility
        self.Rt0   = Rt0
        self.CI    = CI 

        # state and delta vectors 
        if dT0 is None:
            dT0 = np.random.poisson(self.ll) # initial number of new cases 
        self.dT = [dT0] # case change rate, initialized with the first introduction, if any
        self.Rt = [Rt0]
        self.b  = [np.exp(self.gamma * (Rt0 - 1.0))]
        self.S  = [population - E0 - I0 - R0 - D0]
        self.E  = [E0]
        self.I  = [I0] 
        self.R  = [R0]
        self.D  = [D0]
        self.N  = [population - D0] # total population = S + I + R 
        self.beta = [Rt0 * self.gamma] # initial contact rate 
        self.total_cases = [I0] # total cases 
        self.upper_CI = [upper_CI]
        self.lower_CI = [lower_CI]

        np.random.seed(random_seed)
    
    def forward_epi_step(self, dB: int = 0): 
        # get previous state 
        S, E, I, R, D, N = (vector[-1] for vector in (self.S, self.E, self.I, self.R, self.D, self.N))

        # update state 
        Rt = self.Rt0 * float(S)/float(N)
        b  = np.exp(self.gamma * (Rt - 1))

        rate_T    = max(0, self.b[-1] * self.dT[-1])
        num_cases = poisson.rvs(rate_T)
        self.upper_CI.append(poisson.ppf(self.CI,     rate_T))
        self.lower_CI.append(poisson.ppf(1 - self.CI, rate_T))

        E += num_cases
        S -= num_cases

        rate_I    = self.sigma * E
        num_inf   = poisson.rvs(rate_I)

        E -= num_inf 
        I += num_inf

        rate_D    = self.m * self.gamma * I
        num_dead  = poisson.rvs(rate_D)
        D        += num_dead

        rate_R    = (1 - self.m) * self.gamma * I 
        num_recov = poisson.rvs(rate_R)
        R        += num_recov

        I -= (num_dead + num_recov)
        
        if S < 0: S = 0
        if E < 0: E = 0
        if I < 0: I = 0
        if R < 0: R = 0
        if D < 0: D = 0

        N = S + E + I + R
        beta = (num_cases * N)/(b * S * I)

        # update state vectors 
        self.Rt.append(Rt)
        self.b.append(b)
        self.S.append(S)
        self.E.append(E)
        self.I.append(I)
        self.R.append(R)
        self.D.append(D)
        self.N.append(N)
        self.beta.append(beta)
        self.dT.append(num_cases)
        self.total_cases.append(E + I + R + D)

class AR1():
    """ first-order autoregressive model with white noise """
    def __init__(self, phi: float = 1.01, sigma: float = 1, I0: int = 10, random_seed: int = 0):
        self.phi = phi 
        self.sigma = sigma
        self.I = [I0]
        np.random.seed(random_seed)
    
    def set_parameters(self, **kwargs):
        if "phi"   in kwargs: self.phi   = kwargs["phi"]
        if "sigma" in kwargs: self.sigma = kwargs["sigma"]
        return self 

    def run(self, days: int):
        for _ in range(days):
            self.I.append(self.phi * self.I[-1] + np.random.normal(scale = self.sigma))
        return self 

class MigrationSpikeModel(NetworkedSIR):
    """ networked SIR model simulating a population influx at a given time """
    def __init__(self, units: Sequence[SIR], introduction_time: Sequence[int], migratory_influx: Dict[str, int], default_migrations: Optional[np.matrix] = None, random_seed: Optional[int] = None):
        self.counter = 0
        self.migratory_influx  = migratory_influx 
        self.introduction_time = introduction_time
        super().__init__(units, default_migrations, random_seed)

    def tick(self, migrations: np.matrix):
        self.counter += 1
        # run migration step 
        outflux       = [unit.migration_step() for unit in self.units]
        transmissions = [flux * migrations[i, :].sum() for (i, flux) in enumerate(outflux)]
        
        # now run forward epidemiological model, and add spike at intro time 
        if self.counter == self.introduction_time:
            for (unit, tmx) in zip(self.units, transmissions):
                unit.forward_epi_step(tmx + self.migratory_influx[unit.name])
        else: 
            for (unit, tmx) in zip(self.units, transmissions):
                unit.forward_epi_step(tmx)

def gravity_matrix(gdf_path: Path, population_path: Path) -> Tuple[Sequence[str], Sequence[float], np.matrix]:
    gdf = gpd.read_file(gdf_path)
    districts = [d.upper() for d in gdf.district.values]

    pop_df = pd.read_csv(population_path)

    # population count is numeric in Maharashtra data and a string in other data - converting to numeric
    if pop_df["Population(2011 census)"].dtype == object:
        pop_df["Population(2011 census)"] = pop_df["Population(2011 census)"].str.replace(",","").apply(float)

    population_mapping = {k.replace("-", " ").upper(): v for (k, v) in zip(pop_df["Name"], pop_df["Population(2011 census)"])}
    populations = [population_mapping[district.upper()] for district in districts]

    centroids = [list(pt.coords)[0] for pt in gdf.centroid]
    P = distance_matrix(centroids, centroids)
    P[P != 0] = P[P != 0] ** -1.0 
    P *= np.array(populations)[:, None]
    P /= P.sum(axis = 0)

    return (districts, populations, P)
